ALPHANUMERIC_REGEX = r"[a-zA-Z0-9:/\#]+"
# ALPHANUMERIC_ACCENT_REGEX = r"[A-zÀ-ú0-9:/\-\_\#]+"
ALPHANUMERIC_ACCENT_REGEX = r"[A-zÀ-ú0-9]+"
WORDS_ACCENT_REGEX = r"[A-zÀ-ú:/\#-]+"
WORDS_REGEX = r"[a-zA-Z:/\#]+"
URL_REGEX = r"https?://\S+"

CONNECTOR_WORDS = {
    "en": frozenset(
        {
            "a",
            "an",
            "and",
            "at",
            "by",
            "for",
            "from",
            "in",
            "of",
            "on",
            "or",
            "the",
            "to",
            "with",
            "without",
        }
    ),
    "es": frozenset(
        {
            "un",
            "una",
            "y",
            "para",
            "por",
            "en",
            "el",
            "asi",
            "con",
            "si",
            "asi",
            "aunque",
            "de",
            "sino",
            "pero",
            "sin",
            "como",
            "segun",
            "ahora",
            "que",
            "cuando",
            "durante",
            "entonces",
            "hasta",
            "luego",
        }
    ),
}
ROMANCE_SUPPORT = set(
    [
        "fr",
        "fr_BE",
        "fr_CA",
        "fr_FR",
        "wa",
        "frp",
        "oc",
        "ca",
        "rm",
        "lld",
        "fur",
        "lij",
        "lmo",
        "es",
        "es_AR",
        "es_CL",
        "es_CO",
        "es_CR",
        "es_DO",
        "es_EC",
        "es_ES",
        "es_GT",
        "es_HN",
        "es_MX",
        "es_NI",
        "es_PA",
        "es_PE",
        "es_PR",
        "es_SV",
        "es_UY",
        "es_VE",
        "pt",
        "pt_br",
        "pt_BR",
        "pt_PT",
        "gl",
        "lad",
        "an",
        "mwl",
        "it",
        "it_IT",
        "co",
        "nap",
        "scn",
        "vec",
        "sc",
        "ro",
        "la",
    ]
)
